==> polyphony-0.3.0.tar.gz: polyphony-0.3.0/polyphony/compiler/ssa.py <==
﻿from collections import defaultdict, deque
from .dominator import DominatorTreeBuilder, DominanceFrontierBuilder
from .env import env
from .symbol import Symbol
from .ir import *
from .tuple import TupleTransformer
from .usedef import UseDefDetector
from .varreplacer import VarReplacer
from logging import getLogger
logger = getLogger(__name__)


class SSATransformerBase(object):
    def __init__(self):
        pass

    def process(self, scope):
        if scope.is_class() or scope.is_global():
            return
        self.scope = scope
        self.dominance_frontier = {}
        self.usedef = scope.usedef
        self.phis = []

        self._compute_dominance_frontier()
        self._insert_phi()
        self._rename()

        self._remove_useless_phi()
        self._insert_predicate()
        self._cleanup_phi()

    def _sort_phi(self, blk):
        phis = blk.collect_stms(PHI)
        if len(phis) == 1:
            return
        for phi in phis:
            blk.stms.remove(phi)

        phis = sorted(phis, key=lambda p: p.var.qualified_symbol(), reverse=True)
        for phi in phis:
            blk.insert_stm(0, phi)

    def _insert_phi(self):
        phi_symbols = defaultdict(list)
        dfs = set()
        for qsym, def_blocks in self.usedef._def_qsym2blk.items():
            assert isinstance(qsym, tuple)
            if not self._need_rename(qsym[-1], qsym):
                continue
            while def_blocks:
                def_block = def_blocks.pop()
                if def_block not in self.dominance_frontier:
                    continue
                for df in self.dominance_frontier[def_block]:
                    logger.log(0, 'DF of ' + def_block.name + ' = ' + df.name)
                    if qsym in phi_symbols[df]:
                        continue
                    phi_symbols[df].append(qsym)
                    #insert phi to df
                    var = self._qsym_to_var(qsym, Ctx.STORE)
                    if var.is_a(ATTR):
                        assert qsym[0].typ.is_object()
                        var.attr_scope = qsym[0].typ.get_scope()
                    phi = self._new_phi(var, df)
                    df.insert_stm(0, phi)
                    #The phi has the definintion of the variable
                    #so we must add the phi to the df_blocks if needed
                    if qsym not in self.usedef.get_qsyms_defined_at(df):
                        def_blocks.add(df)
                    #this must call after the above checking
                    self._add_phi_var_to_usedef(var, phi)
                    self.phis.append(phi)
                    dfs.add(df)
        # In objectssa, the following code is important to make
        # hierarchical PHI definitions in the proper order.
        for df in dfs:
            self._sort_phi(df)

    def _new_phi(self, var, df):
        phi = PHI(var)
        phi.block = df
        phi.args = [var] * len(df.preds)
        phi.defblks = [None] * len(df.preds)
        phi.lineno = 1
        var.lineno = 1
        return phi

    def _add_phi_var_to_usedef(self, var, phi, is_tail_attr=True):
        if is_tail_attr:
            self.usedef.add_var_def(var, phi)
            if var.is_a(ATTR):
                self._add_phi_var_to_usedef(var.exp, phi, is_tail_attr=False)
        else:
            self.usedef.add_var_use(var, phi)

    def _rename(self):
        qcount = {}
        qstack = {}
        using_vars = set()
        for blk in self.scope.traverse_blocks():
            for var in self.usedef.get_vars_defined_at(blk):
                using_vars.add(var)
            for var in self.usedef.get_vars_used_at(blk):
                using_vars.add(var)
        for var in using_vars:
            key = var.qualified_symbol()
            qcount[key] = 0
            qstack[key] = [(0, None)]

        self.new_syms = set()
        self._rename_rec(self.scope.entry_block, qcount, qstack)

        for var, version in self.new_syms:
            assert var.is_a([TEMP, ATTR])
            if self._need_rename(var.symbol(), var.qualified_symbol()):
                new_name = var.symbol().name + '#' + str(version)
                var_sym = var.symbol()
                new_sym = var_sym.scope.inherit_sym(var_sym, new_name)
                logger.debug(str(new_sym) + ' ancestor is ' + str(var.symbol()))
                var.set_symbol(new_sym)

    def _rename_rec(self, block, count, stack):
        for stm in block.stms:
            if not stm.is_a(PHI):
                for use in self.usedef.get_vars_used_at(stm):
                    assert use.is_a([TEMP, ATTR])
                    key = use.qualified_symbol()
                    i, _ = stack[key][-1]
                    self._add_new_sym(use, i)
            #this loop includes PHI
            for d in self.usedef.get_vars_defined_at(stm):
                #print(stm, d)
                assert d.lineno > 0
                assert d.is_a([TEMP, ATTR])
                key = d.qualified_symbol()
                if self._need_rename(d.symbol(), d.qualified_symbol()):
                    logger.debug('count up ' + str(d) + ' ' + str(stm))
                    count[key] += 1
                i = count[key]
                stack[key].append((i, d))
                self._add_new_sym(d, i)
                if stm.is_a(PHI) and d.is_a(ATTR):
                    self._add_new_sym_rest(d.exp, stack)
        #into successors
        for succ in block.succs:
            #collect phi
            phis = succ.collect_stms(PHI)
            for phi in phis:
                self._add_new_phi_arg(phi, phi.var, stack, block)

        for c in self.tree.get_children_of(block):
            self._rename_rec(c, count, stack)
        for stm in block.stms:
            for d in self.usedef.get_vars_defined_at(stm):
                key = d.qualified_symbol()
                if key in stack and stack[key]:
                    stack[key].pop()

    def _add_new_phi_arg(self, phi, var, stack, block, is_tail_attr=True):
        key = var.qualified_symbol()
        i, v = stack[key][-1]
        if is_tail_attr:
            if i > 0:
                var = var.clone()
                var.ctx = Ctx.LOAD
                var.lineno = v.lineno
                idx = phi.block.preds.index(block)
                phi.args[idx] = var
                phi.defblks[idx] = block
                self._add_new_sym(var, i)
        else:
            self._add_new_sym(var, i)

        if var.is_a(ATTR):
            self._add_new_phi_arg(phi, var.exp, stack, block, is_tail_attr=False)

    def _need_rename(self, sym, qsym):
        return False

    def _add_new_sym(self, var, version):
        assert var.is_a([TEMP, ATTR])
        if self._need_rename(var.symbol(), var.qualified_symbol()):
            self.new_syms.add((var, version))

    def _add_new_sym_rest(self, var, stack):
        assert var.is_a([TEMP, ATTR])
        key = var.qualified_symbol()
        i, _ = stack[key][-1]
        self.new_syms.add((var, i))
        if var.is_a(ATTR):
            self._add_new_sym_rest(var.exp, stack)

    def _qsym_to_var(self, qsym, ctx):
        if len(qsym) == 1:
            return TEMP(qsym[0], ctx)
        else:
            exp = self._qsym_to_var(qsym[:-1], Ctx.LOAD)
            return ATTR(exp, qsym[-1], ctx)

    def dump_df(self):
        for node, dfs in sorted(self.dominance_frontier.items(), key=lambda n: n[0].name):
            logger.debug('DF of ' + node.name + ' is ...' + ', '.join([df.name for df in dfs]))

    def _compute_dominance_frontier(self):
        dtree_builder = DominatorTreeBuilder(self.scope)
        tree = dtree_builder.process()
        tree.dump()
        self.tree = tree

        first_block = self.scope.entry_block
        df_builder = DominanceFrontierBuilder()
        self.dominance_frontier = df_builder.process(first_block, tree)

    def _remove_useless_phi(self):
        udd = UseDefDetector()
        udd.process(self.scope)
        usedef = self.scope.usedef

        def get_sym_if_having_only_1(phi):
            syms = [arg.symbol() for arg in phi.args if arg.is_a([TEMP, ATTR]) and arg.symbol() is not phi.var.symbol()]
            if syms and all(syms[0] is s for s in syms):
                return syms[0]
            else:
                return None
        worklist = deque()
        for blk in self.scope.traverse_blocks():
            worklist.extend(blk.collect_stms(PHI))
        while worklist:
            phi = worklist.popleft()
            if not phi.args:
                #assert False
                logger.debug('remove ' + str(phi))
                phi.block.stms.remove(phi)
                #pass
            else:
                sym = get_sym_if_having_only_1(phi)
                if sym:
                    logger.debug('remove ' + str(phi))
                    if phi in phi.block.stms:
                        phi.block.stms.remove(phi)
                        for a in phi.args:
                            usedef.remove_var_use(a, phi)
                        usedef.remove_var_def(phi.var, phi)
                    replace_var = phi.var.clone()
                    replace_var.set_symbol(sym)
                    replace_var.ctx = Ctx.LOAD
                    replaces = VarReplacer.replace_uses(phi.var, replace_var, usedef)
                    for rep in replaces:
                        if rep.is_a(PHI):
                            worklist.append(rep)

    def _cleanup_phi(self):
        for blk in self.scope.traverse_blocks():
            for phi in blk.collect_stms(PHI):
                remove_args = [arg for arg in phi.args if arg.is_a(TEMP) and arg.symbol() is phi.var.symbol()]
                for arg in remove_args:
                    phi.remove_arg(arg)

    def _concat_predicates(self, predicates, op):
        if not predicates:
            return None
        concat = predicates[0]
        for p in predicates[1:]:
            concat = BINOP(op, concat, p)
        return concat

    def _idom_path(self, blk, path):
        path.append(blk)
        idom = self.tree.get_parent_of(blk)
        if not idom:
            return
        self._idom_path(idom, path)

    def _insert_predicate(self):
        for blk in self.scope.traverse_blocks():
            phis = blk.collect_stms(PHI)
            if not phis:
                continue
            phi_predicates = [pred.path_exp if pred.path_exp else CONST(1) for pred in blk.preds]
            for phi in phis:
                phi.ps = phi_predicates[:]
                assert len(phi.ps) == len(phi.args)


class ScalarSSATransformer(SSATransformerBase):
    def __init__(self):
        super().__init__()

    def _need_rename(self, sym, qsym):
        return not (sym.is_condition() or
                    sym.is_param() or
                    sym.is_return() or
                    sym.typ.name in ['function', 'class', 'object', 'tuple', 'port'])


class TupleSSATransformer(SSATransformerBase):
    def __init__(self):
        super().__init__()

    def process(self, scope):
        if scope.is_class() or scope.is_global():
            return
        super().process(scope)
        UseDefDetector().process(scope)
        TupleTransformer().process(scope)
        UseDefDetector().process(scope)
        self._process_use_phi()

    def _process_use_phi(self):
        usedef = self.scope.usedef
        for blk in self.scope.traverse_blocks():
            phis = blk.collect_stms(PHI)
            for phi in phis:
                uses = usedef.get_stms_using(phi.var.qualified_symbol())
                for use in uses:
                    self._insert_use_phi(phi, use)

    def _insert_use_phi(self, phi, use_stm):
        insert_idx = use_stm.block.stms.index(use_stm)
        use_mrefs = [ir for ir in use_stm.find_irs(MREF) if ir.mem.symbol().typ.is_tuple()]
        qsym = phi.var.qualified_symbol()

        def replace_attr(mref, qsym, newmem):
            if mref.mem.qualified_symbol() == qsym:
                mref.mem = newmem

        for mref in use_mrefs:
            if mref.mem.qualified_symbol() == qsym:
                tmp = self.scope.add_temp('{}_{}'.format(Symbol.temp_prefix,
                                                         mref.mem.symbol().orig_name()))
                var = TEMP(tmp, Ctx.STORE)
                var.lineno = use_stm.lineno
                uphi = UPHI(var)
                uphi.lineno = use_stm.lineno
                uphi.ps = phi.ps[:]
                for arg in phi.args:
                    argmref = mref.clone()
                    argmref.lineno = use_stm.lineno
                    argmref.mem = arg.clone()
                    uphi.args.append(argmref)
                use_stm.block.insert_stm(insert_idx, uphi)
            var = var.clone()
            var.ctx = Ctx.LOAD
            use_stm.replace(mref, var)
        pass

    def _need_rename(self, sym, qsym):
        return sym.typ.is_tuple() and not sym.is_param()


class ObjectSSATransformer(SSATransformerBase):
    def __init__(self):
        super().__init__()

    def process(self, scope):
        if scope.is_class() or scope.is_global():
            return
        super().process(scope)
        self._process_use_phi()

    def _process_use_phi(self):
        usedef = self.scope.usedef
        for blk in self.scope.traverse_blocks():
            phis = blk.collect_stms(PHI)
            for phi in phis:
                uses = usedef.get_stms_using(phi.var.qualified_symbol())
                for use in uses:
                    self._insert_use_phi(phi, use)

    def _insert_use_phi(self, phi, use_stm):
        insert_idx = use_stm.block.stms.index(use_stm)
        use_attrs = [ir for ir in use_stm.kids() if ir.is_a(ATTR)]
        qsym = phi.var.qualified_symbol()

        def replace_attr(attr, qsym, newattr):
            if attr.is_a(ATTR):
                if attr.exp.qualified_symbol() == qsym:
                    attr.exp = newattr
                    return
                return replace_attr(attr.exp, qsym, newattr)
        for use_attr in use_attrs:
            if use_attr.attr.typ.is_object():
                continue
            if use_attr.exp.qualified_symbol() == qsym:
                uphi = UPHI(use_attr.clone())
                uphi.lineno = use_stm.lineno
                uphi.ps = phi.ps[:]
                for arg in phi.args:
                    uarg = use_attr.clone()
                    uarg.lineno = use_stm.lineno
                    replace_attr(uarg, qsym, arg.clone())
                    uphi.args.append(uarg)
                use_stm.block.insert_stm(insert_idx, uphi)

    def _need_rename(self, sym, qsym):
        if not sym.typ.is_object():
            return False
        if sym.name == env.self_name:
            return False
        if sym.scope.is_module() or sym.scope.is_global():
            return False
        if sym.is_param():
            return False
        idx = qsym.index(sym)
        if idx > 0:
            if not self._need_rename(qsym[idx - 1], qsym):
                return False
        return True

==> polyphony-0.3.1.tar.gz: polyphony-0.3.1/polyphony/compiler/ssa.py <==
﻿from collections import defaultdict, deque
from .block import merge_path_exp
from .dominator import DominatorTreeBuilder, DominanceFrontierBuilder
from .env import env
from .symbol import Symbol
from .ir import *
from .tuple import TupleTransformer
from .usedef import UseDefDetector
from .utils import replace_item
from .varreplacer import VarReplacer
from logging import getLogger
logger = getLogger(__name__)


class SSATransformerBase(object):
    def __init__(self):
        pass

    def process(self, scope):
        if scope.is_class() or scope.is_global():
            return
        self.scope = scope
        self.dominance_frontier = {}
        self.usedef = scope.usedef
        self.phis = []

        self._compute_dominance_frontier()
        self._insert_phi()
        self._rename()

        self._remove_useless_phi()
        self._insert_predicate()
        #self._cleanup_phi()
        self._find_loop_phi()

    def _sort_phi(self, blk):
        phis = blk.collect_stms(PHI)
        if len(phis) == 1:
            return
        for phi in phis:
            blk.stms.remove(phi)

        phis = sorted(phis, key=lambda p: p.var.qualified_symbol(), reverse=True)
        for phi in phis:
            blk.insert_stm(0, phi)

    def _insert_phi(self):
        phi_symbols = defaultdict(list)
        dfs = set()
        for qsym, def_blocks in self.usedef._def_qsym2blk.items():
            assert isinstance(qsym, tuple)
            if not self._need_rename(qsym[-1], qsym):
                continue
            while def_blocks:
                def_block = def_blocks.pop()
                if def_block not in self.dominance_frontier:
                    continue
                for df in self.dominance_frontier[def_block]:
                    logger.log(0, 'DF of ' + def_block.name + ' = ' + df.name)
                    if qsym in phi_symbols[df]:
                        continue
                    phi_symbols[df].append(qsym)
                    #insert phi to df
                    var = self._qsym_to_var(qsym, Ctx.STORE)
                    if var.is_a(ATTR):
                        assert qsym[0].typ.is_object()
                        var.attr_scope = qsym[0].typ.get_scope()
                    phi = self._new_phi(var, df)
                    df.insert_stm(0, phi)
                    #The phi has the definintion of the variable
                    #so we must add the phi to the df_blocks if needed
                    if qsym not in self.usedef.get_qsyms_defined_at(df):
                        def_blocks.add(df)
                    #this must call after the above checking
                    self._add_phi_var_to_usedef(var, phi)
                    self.phis.append(phi)
                    dfs.add(df)
        # In objectssa, the following code is important to make
        # hierarchical PHI definitions in the proper order.
        for df in dfs:
            self._sort_phi(df)

    def _new_phi(self, var, df):
        phi = PHI(var)
        phi.block = df
        phi.args = [var] * len(df.preds)
        phi.defblks = [None] * len(df.preds)
        phi.lineno = 1
        var.lineno = 1
        phi.iorder = 0
        var.iorder = 0
        return phi

    def _add_phi_var_to_usedef(self, var, phi, is_tail_attr=True):
        if is_tail_attr:
            self.usedef.add_var_def(var, phi)
            if var.is_a(ATTR):
                self._add_phi_var_to_usedef(var.exp, phi, is_tail_attr=False)
        else:
            self.usedef.add_var_use(var, phi)

    def _rename(self):
        qcount = {}
        qstack = {}
        using_vars = set()
        for blk in self.scope.traverse_blocks():
            for var in self.usedef.get_vars_defined_at(blk):
                using_vars.add(var)
            for var in self.usedef.get_vars_used_at(blk):
                using_vars.add(var)
        for var in using_vars:
            key = var.qualified_symbol()
            qcount[key] = 0
            qstack[key] = [(0, None)]

        self.new_syms = set()
        self._rename_rec(self.scope.entry_block, qcount, qstack)

        for var, version in self.new_syms:
            assert var.is_a([TEMP, ATTR])
            if self._need_rename(var.symbol(), var.qualified_symbol()):
                new_name = var.symbol().name + '#' + str(version)
                var_sym = var.symbol()
                new_sym = var_sym.scope.inherit_sym(var_sym, new_name)
                logger.debug(str(new_sym) + ' ancestor is ' + str(var.symbol()))
                var.set_symbol(new_sym)

    def _rename_rec(self, block, count, stack):
        for stm in block.stms:
            if not stm.is_a(PHI):
                for use in self.usedef.get_vars_used_at(stm):
                    assert use.is_a([TEMP, ATTR])
                    key = use.qualified_symbol()
                    i, _ = stack[key][-1]
                    self._add_new_sym(use, i)
            #this loop includes PHI
            for d in self.usedef.get_vars_defined_at(stm):
                #print(stm, d)
                assert d.lineno > 0
                assert d.is_a([TEMP, ATTR])
                key = d.qualified_symbol()
                if self._need_rename(d.symbol(), d.qualified_symbol()):
                    logger.debug('count up ' + str(d) + ' ' + str(stm))
                    count[key] += 1
                i = count[key]
                assert d.iorder >= 0
                stack[key].append((i, d))
                self._add_new_sym(d, i)
                if stm.is_a(PHI) and d.is_a(ATTR):
                    self._add_new_sym_rest(d.exp, stack)
        #into successors
        for succ in block.succs:
            #collect phi
            phis = succ.collect_stms(PHI)
            for phi in phis:
                self._add_new_phi_arg(phi, phi.var, stack, block)

        for c in self.tree.get_children_of(block):
            self._rename_rec(c, count, stack)
        for stm in block.stms:
            for d in self.usedef.get_vars_defined_at(stm):
                key = d.qualified_symbol()
                if key in stack and stack[key]:
                    stack[key].pop()

    def _add_new_phi_arg(self, phi, var, stack, block, is_tail_attr=True):
        key = var.qualified_symbol()
        i, v = stack[key][-1]
        if is_tail_attr:
            if i > 0:
                var = var.clone()
                var.ctx = Ctx.LOAD
                var.lineno = v.lineno
                var.iorder = v.iorder
                idx = phi.block.preds.index(block)
                phi.args[idx] = var
                phi.defblks[idx] = block
                self._add_new_sym(var, i)
        else:
            self._add_new_sym(var, i)

        if var.is_a(ATTR):
            self._add_new_phi_arg(phi, var.exp, stack, block, is_tail_attr=False)

    def _need_rename(self, sym, qsym):
        return False

    def _add_new_sym(self, var, version):
        assert var.is_a([TEMP, ATTR])
        if self._need_rename(var.symbol(), var.qualified_symbol()):
            self.new_syms.add((var, version))

    def _add_new_sym_rest(self, var, stack):
        assert var.is_a([TEMP, ATTR])
        key = var.qualified_symbol()
        i, _ = stack[key][-1]
        self.new_syms.add((var, i))
        if var.is_a(ATTR):
            self._add_new_sym_rest(var.exp, stack)

    def _qsym_to_var(self, qsym, ctx):
        if len(qsym) == 1:
            return TEMP(qsym[0], ctx)
        else:
            exp = self._qsym_to_var(qsym[:-1], Ctx.LOAD)
            return ATTR(exp, qsym[-1], ctx)

    def dump_df(self):
        for node, dfs in sorted(self.dominance_frontier.items(), key=lambda n: n[0].name):
            logger.debug('DF of ' + node.name + ' is ...' + ', '.join([df.name for df in dfs]))

    def _compute_dominance_frontier(self):
        dtree_builder = DominatorTreeBuilder(self.scope)
        tree = dtree_builder.process()
        tree.dump()
        self.tree = tree

        first_block = self.scope.entry_block
        df_builder = DominanceFrontierBuilder()
        self.dominance_frontier = df_builder.process(first_block, tree)

    def _remove_useless_phi(self):
        logger.debug(str(self.scope))
        udd = UseDefDetector()
        udd.process(self.scope)
        usedef = self.scope.usedef

        def get_sym_if_having_only_1(phi):
            syms = [arg.symbol() for arg in phi.args if arg.is_a([TEMP, ATTR]) and arg.symbol() is not phi.var.symbol()]
            if syms and all(syms[0] == s for s in syms):
                return syms[0]
            else:
                return None
        worklist = deque()
        for blk in self.scope.traverse_blocks():
            worklist.extend(blk.collect_stms(PHIBase))
        while worklist:
            phi = worklist.popleft()
            #if not usedef.get_stms_using(phi.var.symbol()):
            #    if phi in phi.block.stms:
            #        phi.block.stms.remove(phi)
            #    continue
            if not phi.args:
                #assert False
                logger.debug('remove ' + str(phi))
                phi.block.stms.remove(phi)
                #pass
            else:
                sym = get_sym_if_having_only_1(phi)
                if sym:
                    logger.debug('remove ' + str(phi))
                    if phi in phi.block.stms:
                        phi.block.stms.remove(phi)
                        for a in phi.args:
                            usedef.remove_var_use(a, phi)
                        usedef.remove_var_def(phi.var, phi)
                    replace_var = phi.var.clone()
                    replace_var.set_symbol(sym)
                    replace_var.ctx = Ctx.LOAD
                    replaces = VarReplacer.replace_uses(phi.var, replace_var, usedef)
                    for rep in replaces:
                        if rep.is_a(PHI):
                            worklist.append(rep)
                        usedef.remove_use(phi.var, rep)
                        usedef.add_use(replace_var, rep)

    def _cleanup_phi(self):
        for blk in self.scope.traverse_blocks():
            for phi in blk.collect_stms(PHI):
                remove_args = [arg for arg in phi.args if arg.is_a(TEMP) and arg.symbol() is phi.var.symbol()]
                if remove_args:
                    print('!!!', phi, remove_args)
                for arg in remove_args:
                    phi.remove_arg(arg)

    def _concat_predicates(self, predicates, op):
        if not predicates:
            return None
        concat = predicates[0]
        for p in predicates[1:]:
            concat = BINOP(op, concat, p)
        return concat

    def _idom_path(self, blk, path):
        path.append(blk)
        idom = self.tree.get_parent_of(blk)
        if not idom:
            return
        self._idom_path(idom, path)

    def _insert_predicate(self):
        for blk in self.scope.traverse_blocks():
            phis = blk.collect_stms(PHI)
            if not phis:
                continue
            phi_predicates = []
            for pred in blk.preds:
                if len(pred.succs) == 1:
                    p = pred.path_exp if pred.path_exp else CONST(1)
                else:
                    p = merge_path_exp(pred, blk)
                phi_predicates.append(p)

            for phi in phis:
                phi.ps = phi_predicates[:]
                assert len(phi.ps) == len(phi.args)

    def _find_loop_phi(self):
        for blk in self.scope.traverse_blocks():
            phis = blk.collect_stms(PHI)
            if not phis:
                continue
            if not blk.preds_loop:
                continue
            for phi in phis:
                #if phi.ps[0].is_a(CONST) and phi.ps[0].value:
                lphi = LPHI.from_phi(phi)
                replace_item(blk.stms, phi, lphi)


class ScalarSSATransformer(SSATransformerBase):
    def __init__(self):
        super().__init__()

    def _need_rename(self, sym, qsym):
        if (sym.is_condition() or
                sym.is_param() or
                sym.is_return() or
                sym.is_static() or
                sym.typ.name in ['function', 'class', 'object', 'tuple', 'port']):
            return False
        defstms = self.usedef.get_stms_defining(qsym)
        if len(defstms) == 1:
            return False
        return True


class TupleSSATransformer(SSATransformerBase):
    def __init__(self):
        super().__init__()

    def process(self, scope):
        if scope.is_class() or scope.is_global():
            return
        super().process(scope)
        UseDefDetector().process(scope)
        TupleTransformer().process(scope)
        UseDefDetector().process(scope)
        self._process_use_phi()

    def _process_use_phi(self):
        usedef = self.scope.usedef
        for blk in self.scope.traverse_blocks():
            phis = blk.collect_stms(PHI)
            for phi in phis:
                uses = usedef.get_stms_using(phi.var.qualified_symbol())
                for use in uses:
                    self._insert_use_phi(phi, use)

    def _insert_use_phi(self, phi, use_stm):
        insert_idx = use_stm.block.stms.index(use_stm)
        use_mrefs = [ir for ir in use_stm.find_irs(MREF) if ir.mem.symbol().typ.is_tuple()]
        qsym = phi.var.qualified_symbol()

        def replace_attr(mref, qsym, newmem):
            if mref.mem.qualified_symbol() == qsym:
                mref.mem = newmem

        for mref in use_mrefs:
            if mref.mem.qualified_symbol() == qsym:
                tmp = self.scope.add_temp('{}_{}'.format(Symbol.temp_prefix,
                                                         mref.mem.symbol().orig_name()))
                var = TEMP(tmp, Ctx.STORE)
                var.lineno = use_stm.lineno
                var.iorder = use_stm.iorder
                uphi = UPHI(var)
                uphi.lineno = use_stm.lineno
                uphi.iorder = use_stm.iorder
                uphi.ps = phi.ps[:]
                for arg in phi.args:
                    argmref = mref.clone()
                    argmref.lineno = use_stm.lineno
                    argmref.iorder = use_stm.iorder
                    argmref.mem = arg.clone()
                    uphi.args.append(argmref)
                use_stm.block.insert_stm(insert_idx, uphi)
            var = var.clone()
            var.ctx = Ctx.LOAD
            use_stm.replace(mref, var)
        pass

    def _need_rename(self, sym, qsym):
        return sym.typ.is_tuple() and not sym.is_param()


class ObjectSSATransformer(SSATransformerBase):
    def __init__(self):
        super().__init__()

    def process(self, scope):
        if scope.is_class() or scope.is_global():
            return
        super().process(scope)
        self._process_use_phi()

    def _process_use_phi(self):
        usedef = self.scope.usedef
        for blk in self.scope.traverse_blocks():
            phis = blk.collect_stms(PHI)
            for phi in phis:
                uses = usedef.get_stms_using(phi.var.qualified_symbol())
                for use in uses:
                    self._insert_use_phi(phi, use)

    def _insert_use_phi(self, phi, use_stm):
        insert_idx = use_stm.block.stms.index(use_stm)
        use_attrs = [ir for ir in use_stm.kids() if ir.is_a(ATTR)]
        qsym = phi.var.qualified_symbol()

        def replace_attr(attr, qsym, newattr):
            if attr.is_a(ATTR):
                if attr.exp.qualified_symbol() == qsym:
                    attr.exp = newattr
                    return
                return replace_attr(attr.exp, qsym, newattr)
        for use_attr in use_attrs:
            if use_attr.attr.typ.is_object():
                continue
            if use_attr.exp.qualified_symbol() == qsym:
                uphi = UPHI(use_attr.clone())
                uphi.lineno = use_stm.lineno
                uphi.iorder = use_stm.iorder
                uphi.ps = phi.ps[:]
                for arg in phi.args:
                    uarg = use_attr.clone()
                    uarg.lineno = use_stm.lineno
                    uarg.iorder = use_stm.iorder
                    replace_attr(uarg, qsym, arg.clone())
                    uphi.args.append(uarg)
                use_stm.block.insert_stm(insert_idx, uphi)

    def _need_rename(self, sym, qsym):
        if not sym.typ.is_object():
            return False
        if sym.name == env.self_name:
            return False
        if sym.scope.is_module() or sym.scope.is_global():
            return False
        if sym.is_param():
            return False
        idx = qsym.index(sym)
        if idx > 0:
            if not self._need_rename(qsym[idx - 1], qsym):
                return False
        return True
